###########################################################################
# Python-Programm für k-Means-Clustering zur Klassifizierung
#
# Als Testbeispiel soll das Programm lernen, ob ein Punkt P(x1|x2) zu einem
# von 3 Clustern gehört, Die Anzahl der Cluster (3) ist vorab bekannt.
#
##########################################################################
import numpy as np               # stellt Mathe-Hilfsprogramme bereit
import matplotlib.pyplot as plt  # stellt Hilfsprogramme für Diagramme bereit
import random                    # stellt Hilfsprogramm für Zufallszahlen bereit

random.seed(0) 

# Funktionsunterprogramm für Zufallszahlen zwischen -1 und +1
def zuza():
    zufall=random.random()          # Zufallszahlen zwischen 0 und 1
    zz=2.*zufall-1.                 # Zufallszahlen zwischen -1 und 1
    return zz


#########################################################################
# Hier beginnt das Programm zum Finden der Cluster
#########################################################################

# Startwerte für 3 Zentren der zufälligen Daten

C10=([0.8,0.1])
C20=([-0.3,0.6])
C30=([-0.4,-0.4])

nc1=20              # Anzahl der Daten um Zentrum 1 
nc2=15              # Anzahl der Daten um Zentrum 2
nc3=30              # Anzahl der Daten um Zentrum 3
nges=nc1+nc2+nc3    # Gesamtzahl der Daten
nit=10              # Anzahl der Korrekturschritte für die Zentroiden 

# Vereinbarung von Vektoren für die Ergebnisbilder

C1,C2,C3=[np.zeros((nges,2)) for i in range(3)]        # Daten vorher
C1F,C2F,C3F=[np.zeros((nges,2)) for i in range(3)]     # Daten nachher
XD=np.zeros((nges,2))                                  # alle Daten
XZ=np.zeros(nges)                                      # Zuweisungslabel 
Z1,Z2,Z3=[np.zeros(2) for i in range(3)]               # Zentroidenkoordinaten
WZ1,WZ2,WZ3=[np.zeros((nit,2)) for i in range(3)]      # zentroidenwege
A,IZ=[np.zeros(3) for i in range(2)]                   # Abstandsvektor

# Erzeugung zufälliger Punkte um die Zentren
k=-1
for i in range(0,nc1):
    C1[i,0]=C10[0]+zuza()*0.5
    C1[i,1]=C10[1]+zuza()*0.6
    k=k+1
    XD[k,:]=C1[i,:]

for i in range(0,nc2):
    C2[i,0]=C20[0]+zuza()*0.3
    C2[i,1]=C20[1]+zuza()*0.4
    k=k+1
    XD[k,:]=C2[i,:]

for i in range(0,nc3):
    C3[i,0]=C30[0]+zuza()*0.5
    C3[i,1]=C30[1]+zuza()*0.4
    k=k+1
    XD[k,:]=C3[i,:]

# Zeichnen der Daten vor dem Clustering
plt.figure(1)
plt.subplot(111,aspect='equal')
plt.plot(C1[0:nc1,0],C1[0:nc1,1],"r.")
plt.plot(C2[0:nc2,0],C2[0:nc2,1],"g.")
plt.plot(C3[0:nc3,0],C3[0:nc3,1],"b.")
plt.grid(True)
plt.xlabel("x1")                                      # x-Achsen-Beschriftung
plt.ylabel("x2")                                      # y-Achsen-Beschriftung
plt.title("Ausgangsdaten mit vorgegebenen Clustern")
plt.tight_layout(pad=1)

# Startwerte für Zentroiden

Z1=[-1,-0.8]
Z2=[zuza(),zuza()]
Z3=[zuza(),zuza()]

WZ1[0,:]=Z1        # Speichern der Startwerte für die Wegkurve             "
WZ2[0,:]=Z2
WZ3[0,:]=Z3

plt.figure(2)
ax=plt.subplot(111,aspect='equal')
plt.grid(True)
plt.plot(WZ1[0,0],WZ1[0,1],"r*")    
plt.plot(WZ2[0,0],WZ2[0,1],"g*")
plt.plot(WZ3[0,0],WZ3[0,1],"b*")

for j in range(1,nit):

    # Berechnung der Abstände
   
    for k in range(0,nges-1):
        A[0]=((XD[k,:]-Z1)@(XD[k,:]-Z1))**0.5
        A[1]=((XD[k,:]-Z2)@(XD[k,:]-Z2))**0.5
        A[2]=((XD[k,:]-Z3)@(XD[k,:]-Z3))**0.5
        IZ=np.argsort(A)     # Zuordnung zum nächsten Zentroiden
        XZ[k]=IZ[0]

    Z1,Z2,Z3=[np.zeros(2) for i in range(3)]

    ic1=0
    ic2=0
    ic3=0

    # Sumation der Koordinaten der den Zentroiden zugeordenten Datenpunkte    
   
    for k in range(0,nges):
        if(XZ[k]==0):
            Z1=Z1+XD[k,:]
            C1F[ic1,:]=XD[k,:]
            ic1=ic1+1
        elif(XZ[k]==1):    
            Z2=Z2+XD[k,:]
            C2F[ic2,:]=XD[k,:]           
            ic2=ic2+1
        elif(XZ[k]==2):    
            Z3=Z3+XD[k,:]
            C3F[ic3,:]=XD[k,:]           
            ic3=ic3+1

    # Neue Zentroidkoordinaten
    
    if(ic1>0):
        Z1=Z1/ic1
        WZ1[j,:]=Z1
##    else:
##        WZ1[j,:]=Z1
        
    if(ic2>0):
        Z2=Z2/ic2
        WZ2[j,:]=Z2
##    else:
##        WZ2[j,:]=Z2
        
    if(ic3>0):
        Z3=Z3/ic3
        WZ3[j,:]=Z3
##    else:
##        WZ3[j,:]=Z3

# Zeichnen der Endpunkte der Zentroiden 

plt.plot(Z1[0],Z1[1],"ro")     
plt.plot(Z2[0],Z2[1],"go")
plt.plot(Z3[0],Z3[1],"bo")

# Zeichnen der gefundenen Cluster

plt.plot(C1F[0:ic1,0],C1F[0:ic1,1],"rx",label="Cluster 1")
plt.plot(C2F[0:ic2,0],C2F[0:ic2,1],"gx",label="Cluster 2")
plt.plot(C3F[0:ic3,0],C3F[0:ic3,1],"bx",label="Cluster 3")

# Zeichnen der Zentroidenwege

plt.plot(WZ1[:,0],WZ1[:,1],"--r.",label="Zentroidenweg 1")    
plt.plot(WZ2[:,0],WZ2[:,1],"--g.",label="Zentroidenweg 2")
plt.plot(WZ3[:,0],WZ3[:,1],"--b.",label="Zentroidenweg 3")

plt.xlabel("x1")                                      # x-Achsen-Beschriftung
plt.ylabel("x2")                                      # y-Achsen-Beschriftung
plt.title("Daten nach K-Means-Clustering")
plt.tight_layout(pad=1)
box= ax.get_position()
ax.set_position([box.x0,box.y0,box.width*0.6,box.height])
ax.legend(facecolor='white',bbox_to_anchor=(1.05,1.0),loc='upper left')
plt.show()
 
