###########################################################################
# Berechnung der Ereignisbaumverzweigungen für das Fussball-Wetterbeispiel
# mit Entropie- und Gini-Index-Kriterien
###########################################################################
# Importiere die notwendigen Bibliotheken
import numpy as np
import matplotlib.pyplot as plt
import sys

# Hilfsprogramme

# Zweierlogarithmus
def log2(x):
    y=np.log10(x)/np.log10(2)
    return y

# Entropieberechnung für Skalar
def ent(p):
    if(p==0):        # Ausnahme 0*log(0) :=0
        aux=0
    elif(1-p==0):    # Ausnahme 0*log(0) :=0  
        aux=0
    else:
        aux=-p*log2(p)
    return aux

# Entropieberechnung für Vektor
def entvec(p):
    m=len(p)
    aux1=np.zeros(m)
    for i in range(0,m):
        if(p[i]==0):  # Ausnahme 0*log(0) :=0
            aux1[i]=0
        else:
            aux1[i]=p[i]*log2(p[i])
    aux2=-np.sum(aux1)
    return aux2

# Berechnung der Gewinnfunktionen für Gini- und Entropiekriterium
def gewinn(nsp,nt):

    for i in range(0,nsp):
        sj=float(np.sum(d[:,i]))       # gesamte Jas in Spalte i
        sn=nt-sj                       # gesamte Neins in Spalte i
    
        a[:,i]=d[:,i]*d[:,nsp-1]       # Verknüpfung Spalte d_i und Ergebnisspalte zur Feststellung richtiger Zuordnungen für Ziel-Ja aus den Jas von Spalte i
        nj=np.sum(a[:,i])              # Summe der korrekten Richtig-Zuordnungen für Ziel-Ja aus den Jas von Spalte i 
        nn=sj-nj                       # Summe falscher Zuordnungen für Ziel-Jas aus den Jas von Spalte i
   
        pcj=nj/(sj+1.e-12)             # Wahrscheinlichkeit für korrekte Zuordnung der positiven Merkmale aus Spalte i zu posotiven Ergebnissen
        gja=2.*pcj*(1.-pcj)            # zugehöriger Gini-Index  
        eja=ent(pcj)                   # zugehörige Entropie
    
        b[:,i]=(1-d[:,i])*(1-d[:,nsp-1])  # Verknüpfung nicht(Spalte d_i) und nicht(Ergebnisspalte) zeigt gleiche und ungleiche Ergebnisse bei falschen Zuordnungen
        nnj=np.sum(b[:,i])                # Summe der korrekten Falsch-Zuordnungen für Ziel-Nein aus den Neins von Spalte i
        nnn=sn-nnj                        # Summe der falschen Falsch-Zuordnungen

        pcn=nnj/(nt-sj+1.e-12)            # Wahrscheinlichkeit für korrekte Negativzuordnungen aus Spalte i
        gne=2.*pcn*(1.-pcn)               # zugehöriger Gini-Index
        ene=ent(pcn)                      # zugehörige Entropie
    
        gsp[i]=sj/nt*gja+(nt-sj)/nt*gne   # gewichtetes Mittel der Gini-Indices
        esp[i]=sj/nt*eja+(nt-sj)/nt*ene   # gewichtetes Mittel der Entropie

        if(i<nsp-1):                      # Ausdrucken ohne Ergebnisspalte von Gini-Gewinn und Informationsgewinn
            print('|{:6}|{:11}|{:17}|'.format(i,round(eges-esp[i],3),round(gges-gsp[i],3)))
    return esp,gsp 

# Hauptprogramm ##################################################################################

# Erstelle ein Beispiel-Array (Tabelle) für Training
data = [
    [20, 65, 10, 0, 1],  # Schönes Wetter, wir spielen Fußball
    [25, 80, 5, 1, 0],   # Es regnet, wir spielen nicht
    [18, 70, 15, 0, 1],  # Noch gutes Wetter, wir spielen
    [10, 90, 20, 1, 0],  # Kalt und regnerisch, wir spielen nicht
    [15, 85, 25, 0, 0],  # Zu windig, wir spielen nicht
    [22, 60, 5, 0, 1],   # Schönes Wetter, wir spielen
    [30, 50, 10, 0, 0],  # Zu heiß, wir spielen nicht

# nicht verwendete Datensätze
#   [5, 30, 1, 0, 1],    # Neu: kalt aber trocken, wir spielen     
#   [20, 70, 30, 0, 0],  # Neu: Schönes Wetter, aber zu windig, wir spielen  nicht

    [20, 70, 8, 0, 1]   # Schönes Wetter, wir spielen    
]

# Trenne die Merkmale (Features) und das Ziel (Target)
X = [row[:-1] for row in data]
y = [row[-1] for row in data]

#################################################################
# 1. Entscheidungsebene #########################################

# Gini-Index und Entropie des gesamten Datensatzes

p=([0.5,0.5])
gges=1.-(p[0])**2-(p[1])**2
print('   ')
print("Gini-Index des gesamten Datensatzes = "+str(gges))
eges=entvec(p)
print("Entropie des gesamten Datensatzes = "+str(eges))
print('   ')
#sys.exit()

# Benötigte Felder

gsp=np.zeros(19)
esp=np.zeros(19)
nt=8
a=np.zeros([nt,19])
b=np.zeros([nt,19])
d=np.zeros([nt,19])

# Merkmalsmatrix

d=np.array([[1,1,1,0,0,0,1,1,0,0,0,0,1,1,0,0,0,0,1],
           [1,1,1,1,1,0,1,1,1,1,0,0,0,0,0,0,0,1,0],
           [1,1,0,0,0,0,1,1,1,0,0,0,1,1,1,0,0,0,1],
           [0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,1,0],
           [1,0,0,0,0,0,1,1,1,1,1,0,1,1,1,1,1,0,0],
           [1,1,1,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1],
           [1,1,1,1,1,1,0,0,0,0,0,0,1,1,0,0,0,0,0],
           [1,1,1,0,0,0,1,1,1,0,0,0,1,0,0,0,0,0,1]])
print('   ')
print('| |  0|  1|  2|  3|  4|  5|  6|  7|  8|  9| 10| 11| 12| 13| 14| 15| 16| 17|| 18|')
print('|------------------------------------------------------------------------------|')
print('| |<T1|<T2|<T3|<T4|<T5|<T6|<H1|<H2|<H3|<H4|<H5|<H6|<V1|<V2|<V3|<V4|<V5|  R||  E|')
print('|------------------------------------------------------------------------------|')
i=0
for row in d:
    i=i+1
    print('|{:1}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}||{:3}|'.format(i,*row))
#sys.exit()
print('   ')    
print('|Spalte| I(E)-G(B) | gini(E)-GINI(B) |')
print('|------------------------------------|')

# Berechnung der Gewinnfunktionen

nsp=19
esp,gsp=gewinn(nsp,nt)

# Ergebnis für 1. Ebene

print('   ')
print("1. Entscheidungsebene:")
args=np.argsort(eges-esp[0:nsp-1])    # Sortieren nach Index aller Spalten ohne letzte Ergebnisspalte
print("Maximaler Informationsgewinn in Spalte "+str(args[nsp-2])+" = "+str(round(np.max(eges-esp[0:nsp-2]),3) ) )
print("Maximaler Ginigewinn in Spalte "+str(args[nsp-2])+" = "+str(round(np.max(gges-gsp[0:nsp-2]),3) ) )
print('   ')
print('----------------------------------------------------------')
print('   ')
#sys.exit()

############################################################
# 2. Entscheidungsebene ####################################

p=([0.8,0.2])
gges=1.-(p[0])**2-(p[1])**2
print("Gini-Index des reduzierten Datensatzes im Ja-Zweig = "+str(round(gges,3)))
eges=entvec(p)
print("Entropie des reduzierten Datensatzes im Ja-Zweig = "+str(round(eges,3)))
print('   ')

# Benötigte Felder

esp=np.zeros(16)
gsp=np.zeros(16)
nt=5
a=np.zeros([nt,16])
b=np.zeros([nt,16])
d=np.zeros([nt,16])
d=np.array([[1,1,1,0,0,0,1,1,0,1,1,0,0,0,0,1],
           [1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,1],
           [1,1,1,1,0,0,1,0,0,0,0,0,0,0,0,1],
           [1,1,1,1,1,1,0,0,0,1,1,0,0,0,0,0],
           [1,1,1,0,0,0,1,1,1,1,0,0,0,0,0,1]])
print('| |  0|  1|  2|  3|  4|  5|  6|  7|  8|  9| 10| 11| 12| 13| 14|| 15|')
print('|------------------------------------------------------------------|')
print('| |<T1|<T2|<T3|<T4|<T5|<T6|<H1|<H2|<H3|<V1|<V2|<V3|<V4|<V5|  R||  E|')
print('|------------------------------------------------------------------|')
i=0
for row in d:
    i=i+1
    print('|{:1}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}|{:3}||{:3}|'.format(i,*row))
print("     ")
print('|Spalte| I(E)-G(B) | gini(E)-GINI(B) |')
print('|------------------------------------|')

# Berechnung der Gewinnfunktionen

nsp=16
esp,gsp=gewinn(nsp,nt)

# Ergebnis für 2. Ebene

print('   ')
print("2. Entscheidungsebene:")
args=np.argsort(eges-esp[0:15])
print("Maximaler Informationsgewinn in Spalte "+str(args[14])+" = "+str(round(np.max(eges-esp[0:15]),4) ) )
print("Maximaler Ginigewinn in Spalte "+str(args[14])+" = "+str(round(np.max(gges-gsp[0:15]),4) ) )
print("Spalten 4 und 6 sind ebenfalls möglich")
print('   ')


