###################################
# KNN-Algorithmus mit k=5
###################################
import numpy as np               # stellt Mathe-Hilfsprogramme bereit
import matplotlib.pyplot as plt  # Hilfsprogramme für Diagramme
import random                    # gibt Zufallszahlen

random.seed(0)

# Funktionsunterprogramm für Zufallszahlen zwischen -1 und 1
def zuza():
    zufall=random.random()
    zz=2.*zufall-1
    return zz

####################################################################
# Hier beginnt das Programm für die Durchführung des KNN-Algorithmus
####################################################################

# Startwerte für die Rechnung
ntr=500     # Anzahl der Trainingsdaten
nte=50      # Anzahl der Testdaten
k=2         # Steigung der Grenzkurve
d=-0.5      # Achsenabschnitt der Grenzkurve
n=2         # Grad des Polynoms

X1=np.zeros(ntr)  # Vektor für x-Werte der Trainingsdaten (alle)
X2=np.zeros(ntr)  # Vektor für y-Werte der Trainingsdaten (alle)

X1o=np.zeros(ntr)  # Vektor für x-Werte der Trainingsdaten (oberhalb)
X2o=np.zeros(ntr)  # Vektor für y-Werte der Trainingsdaten

X1u=np.zeros(ntr)  # Vektor für x-Werte der Trainingsdaten (unterhalb)
X2u=np.zeros(ntr)  # Vektor für y-Werte der Trainingsdaten

T=np.zeros(ntr)
abstand=np.zeros(ntr)

io=0
iu=0

# Schleife für Erzeugung der Trainingsdaten

for i in range(0,ntr):
    X1[i]=zuza()                 # Erzeugen eines zuf"alligen x1-Wertes und Speichern im Feld X1
    X2[i]=zuza()                 # Erzeugen eines zuf"alligen x2-Wertes und Speichern im Feld X2

    if(X2[i] > k*X1[i]**n+d):    # Prüfung auf Lage des Trainingspunktes relativ zur Grenzgeraden / Punkt ist oberhalb 
        T[i]=1                   # Kennzeichnung 1
        X1o[io]=X1[i]            # Speichern in den Feldern X1o, X2o
        X2o[io]=X2[i]
        io=io+1                  # Zählen der Punkte oberhalb
    else:                        # Punkt ist unterhalb
        T[i]=0                   # Kennzeichnung 0
        X1u[iu]=X1[i]            # Speichern in den Feldern X1u, X2u
        X2u[iu]=X2[i]
        iu=iu+1                  # Zählen der Punkte unterhalb

# Grenzkurve
X1G=np.linspace(-1,1,100)
X2G=k*X1G**n+d

# Diagramme

plt.figure()                       # <------Beginnt ein Bild
ax=plt.subplot(111)                # <------Handle für ein Teildiagramm (hier insgesamt nur ein Bild) 
plt.plot(X1o[0:io],X2o[0:io],"rx") # Zeichnet Trainingsdaten über Grenzkurve
plt.plot(X1u[0:iu],X2u[0:iu],"b+") # Zeichnet Trainingsdaten unter Grenzkurve
plt.plot(X1G,X2G,)                 # Zeichnet die Grenzkurve
plt.ylim(-1,1)                     # Begrenzt x2-Achse auf Intervall [-1,1]
plt.grid(True)                     # Zeichnet ein Gitternetz 


# Testen mit zufälligen Testdaten

for j in range(0,nte):          # Schleife für alle Testdaten 
    x1t=zuza()                  # zufälliger x1-Wert für Testpunkt 
    x2t=zuza()                  # zufälliger x2-Wert für Testpunkt

# Bestimmung der Abstände zwischen Testpunkt und Trainingsdaten

    for i in range(0,ntr):      # Schleife über alle Trainingsdaten 
        a=x1t-X1[i]             # x1-Koordinatenabstand
        b=x2t-X2[i]             # x2-Koordinatenabstand
        abstand[i]=(a**2+b**2)**(0.5)   # Abstandsberechnung mit Pythagoras

    numsort=np.argsort(abstand)  # Sortieren der Abstände nach Punktnummern

    nk=5                         # Anzahl der Nachbarn für Prüfung
    x1n=np.zeros(2*nk)           # x1- und x2- Koordinaten fürs Zeichnen der Nachbarschaften
    x2n=np.zeros(2*nk)

    tsumme=0
    for k in range(0,nk):        # Kennzeichnungen der 5 nächsten Nachbarn werden summiert
        ii=numsort[k]            # Merken der Nummern der 5 nächsten Punkte
        tsumme=tsumme+T[ii]      # Summieren der Kennzeichnungen der Nachbarn 
        x1n[2*k]=X1[ii]          # Abwechselndes Speichern von nächsten Nachbarn und Testpunktdaten
        x2n[2*k]=X2[ii]
        x1n[2*k+1]=x1t
        x2n[2*k+1]=x2t

    ttest=tsumme/nk              # Mittelwert der 5 Kennzeichnungen
    if(ttest>0.5):               # Entscheidung ob man näher bei 1 (oberhalb) oder 0 (unterhalb) ist
        plt.plot(x1t,x2t,"mo")   # Zeichne Testpunkt mit Farbe magenta für oberhalb
        plt.plot(x1n,x2n,":")
    else:
        plt.plot(x1t,x2t,"co")   # Zeichne Testpunkt mit Farbe cyan für unterhalb
        plt.plot(x1n,x2n,":")

plt.xlabel("x1-Koordinate")
plt.ylabel("x2-Koordinate")
plt.title("K-Nearest-Neighbour-Algorithmus mit k=5")
# Vorbereitung für Legende außerhalb des Diagramms
box= ax.get_position()                                      # Holen der Boxgröße, in der das Diagramm gezeichnet wird       
ax.set_position([box.x0,box.y0,box.width*0.75,box.height])  # Reduktion der Breite auf 75% damit Platz für Legende ist 
# Zeichnet Legende rechts oben außerhalb des Diagramms
plt.legend(labels=["Training","Training","Grenzkurve", "Testen","Nachbarn","Testen"],bbox_to_anchor=(1.05,1.0),loc='upper left')
plt.show()

