###########################################################################
# Python-Programm für logistische Regression zur Klassifizierung
#
# Als Testbeispiel soll das Programm lernen, ob ein Punkt P(x1|x2) oberhalb oder
# unterhalb einer Grenzkurve x2 = k*x1**n +d mit n=1 oder 2 liegt mit x1 und y1-Werten
# im Intervall [-1,1].
#
##########################################################################
import numpy as np               # stellt Mathe-Hilfsprogramme bereit
import matplotlib.pyplot as plt  # stellt Hilfsprogramme für Diagramme bereit
import random                    # stellt Hilfsprogramm für Zufallszahlen bereit
from mpl_toolkits import mplot3d

random.seed(0) 

# Funktionsunterprogramm für Zufallszahlen zwischen -1 und +1
def zuza():
    zufall=random.random()          # Zufallszahlen zwischen 0 und 1
    zz=2.*zufall-1.                 # Zufallszahlen zwischen -1 und 1
    return zz

# Funktionsunterprogramm für Sigmoidfunktion
def lofu(w1,w2,b,x1,x2):
    z=1./(1.+np.exp(-(w1*x1+w2*x2+b))) 
    return z

# Funktionsunterprogramm für Grenzkurve zwischen den Gebieten, wo Punkte klassifiziert werden
def grenze(ig,x):
    if(ig==1):      
        y=k*x+d                     # Grenzkurve ist Gerade
    elif(ig==2):
        y=k*x**2+d                  # Grenzkurve ist Parabel
    return y

#########################################################################
# Hier beginnt das Programm für Training und Testen des neuronalen Netzes
#########################################################################

# Parameter und Startwerte für die Rechnung
ntr=500                   # Anzahl der Trainingsdatensätze
nte=150                   # Anzahl der Testdatensätze
nga=400                   # Anzahl der Schritte für Gradientenanstiegsverfahren
eta=0.0001 #*10               # Lernrate
ig=1                     # Auswahl der Grenzkurve (ig=1: Gerade, ig=2: Parabel)
d=0.7#-0.5                    # Achsenabschnitt der Begrenzungsgerade oder y-Verschiebung der Parabel
k=-2#2                       # Steigung der Begrenzungsgerade oder Koeffizient vo x**2

# Bereitstellung der Felder mit Länge ntr, die später gebraucht werden 

Z1,Z2,Z3=[np.zeros(2) for i in range(3)]               # Zentroidenkoordinaten
X1,X2,T,X1p,X2p,X1m,X2m,X1tp,X2tp,X1tm,X2tm=[np.zeros(ntr) for i in range(11)]
IG,W1,W2,B=[np.zeros(nga) for i in range(4)]

ip=0
im=0
ipt=0
imt=0
inga=0

lolifu=0.
dlw1=0.
dlw2=0.
dlb=0.

w1=zuza()*0.5
w2=zuza()*0.5
b=zuza()*0.5

W1[0]=-w1/w2
W2[0]=-b/w2
print("=Startwerte================================")
print(w1,w2,b)
print(-w1/w2,-b/w2)


# Schleife für Trainingsdaten und Ableitungen der Zielfunktion
for i in range(0,ntr):

# Erstelle einen neuen Trainingsdatensatz (x1,x2,t)

    x1=zuza()              # gibt zufälligen Wert für x1 zwischen -1 und 1 
    x2=zuza()              # gibt zufälligen Wert für x2 zwischen -1 und 1 
    if(x2>grenze(ig,x1)):  # prüft, ob x2 oberhalb der Grenzkurve liegt (ig=1: Gerade oder ig=1: Parabel)        
        t=1.               # wenn ja, wird t=1 gesetzt
        X1m[im]=x1
        X2m[im]=x2
        im=im+1   
    else:
        t=0.                # wenn nein, wird t=0 gesetzt
        X1p[ip]=x1
        X2p[ip]=x2
        ip=ip+1
    X1[i]=x1
    X2[i]=x2
    T[i]=t

# Gradientenabstieg bzw. -aufstieg

for inga in range(1,nga):

    # Schleife für Trainingsdaten und Ableitungen der Zielfunktion
    for i in range(0,ntr):

    # Erstelle einen neuen Trainingsdatensatz (x1,x2,t)

        x1=zuza()              # gibt zufälligen Wert für x1 zwischen -1 und 1 
        x2=zuza()              # gibt zufälligen Wert für x2 zwischen -1 und 1 
        if(x2>grenze(ig,x1)):  # prüft, ob x2 oberhalb der Grenzkurve liegt (ig=1: Gerade oder ig=1: Parabel)        
            t=1.               # wenn ja, wird t=1 gesetzt
        else:
            t=0.                # wenn nein, wird t=0 gesetzt

    # Zielfunktion Maximum Likelihood und Ableitungen
        sig=lofu(w1,w2,b,x1,x2)
        if(1.-sig<=0):
            sig=0.9999999
        lolifu=lolifu+t*np.log(sig)+(1.-t)*np.log(1.-sig)
        dlw1=dlw1+(t-sig)*x1
        dlw2=dlw2+(t-sig)*x2
        dlb=dlb+(t-sig)

# Korrekturen
    w1=w1+eta*dlw1
    w2=w2+eta*dlw2
    b =b +eta*dlb
    IG[inga]=inga           # Speichern der Verlaufs der Korrekturen 
    W1[inga]=-w1/w2
    W2[inga]=-b/w2
    B[inga]=b

print("=Endwerte================================")
print(w1,w2,b)
print("=Endwerte bezogen auf w2 ==============================")
print(-w1/w2,1.,-b/w2)

# Berechnen der Grenzkurve

x1g=np.linspace(-1,1,101)
x2g=grenze(ig,x1g)

# Zeichnen der Trainingsdaten und der Grenzkurve

plt.figure(1)#figsize=(6,4))
ax=plt.subplot(111)
plt.plot(X1m[0:im],X2m[0:im],"r.")  # Trainingsdaten über Grenzkurve
plt.plot(X1p[0:ip],X2p[0:ip],"b.",label='Training')  # Trainingsdaten unter Grenzkurve
plt.plot(x1g,x2g,"g",label='Grenzkurve')             # Zeichne Grenzkurve

# Testen mit nte zufällig gewählten Testdaten

for j in range(1,nte):

    x1t=zuza()              # gibt zufälligen Wert für x1 zwischen -1 und 1 
    x2t=zuza()              # gibt zufälligen Wert für x2 zwischen -1 und 1
    
    sig=lofu(w1,w2,b,x1t,x2t)
    if (sig>0.5):                # Speichere Testdatenpunkte oberhalb ...
        ipt=ipt+1
        X1tp[ipt]=x1t     
        X2tp[ipt]=x2t
    else:                        # ... und unterhalb der Grenzkurve
        imt=imt+1
        X1tm[imt]=x1t
        X2tm[imt]=x2t

plt.plot(X1tp[1:ipt],X2tp[1:ipt],"mo")         # Zeichne Testpunkt, die oberhalb erkannt wurden
plt.plot(X1tm[1:imt],X2tm[1:imt],"co")         # Zeichne Testpunkte, die unterhalb erkannt wurden
plt.ylim(-1,1)                   # Begrenze y-Achse im Diamgramm zwischen -1 und 1
plt.grid(True)                   # Zeichne Gitterlinien
plt.xlabel("x1")                 # x-Achsen-Beschriftung
plt.ylabel("x2")                 # y-Achsen-Beschriftung
plt.title("Logistische Regression")    # Überschrift
# Verkürzung des Bildes in x-Richtung und Zeichnen der Legende außerhalb 
box= ax.get_position()
ax.set_position([box.x0,box.y0,box.width*0.75,box.height])
ax.legend(labels=["Training","Training","Grenzkurve", "Testen","Testen"],facecolor='white',bbox_to_anchor=(1.05,1.0),loc='upper left') 
plt.show()

plt.figure(2)
ax=plt.subplot(111)
plt.plot(IG,W1,"r")
plt.plot(IG,W2,"g")
plt.grid(True)                   # Zeichne Gitterlinien
plt.xlabel("Schrittnummer Gradientenabstieg")                 # x-Achsen-Beschriftung
plt.ylabel("k,d")                 # y-Achsen-Beschriftung
plt.title("Entwicklung von k und d beim Gradientenabstieg")
# Verkürzung des Bildes in x-Richtung und Zeichnen der Legende außerhalb
plt.tight_layout(pad=1)
box= ax.get_position()
ax.set_position([box.x0,box.y0,box.width*0.6,box.height])
ax.legend(labels=["Steigung k","Achsenabschnitt d"],facecolor='white',bbox_to_anchor=(1.05,1.0),loc='upper left')


plt.show()                        # Zeige Ergebnisbild am Bildschirm





 

   
