import numpy as np
import matplotlib.pyplot as plt

# Daten Prüfungsbeispiel

ng=29      # Prüfungsteilnehmer
nb=22      # Anzahl bestanden
nd=ng-nb   # Anzahl nicht bestanden

ujb=21     # am Unterricht immer teilgenommen und bestanden
ujd=4      # am Unterricht immer teilgenommen und nicht bestanden 
unb=nb-ujb # nicht immmer im Unterricht und bestanden
und=nd-ujd # nicht immer im Unterricht und nicht bestanden

sjb=16     # Stoff geübt und bestanden
sjd=2      # Stoff geübt und nicht bestanden
snb=nb-sjb # Stoff nicht geübt und bestanden
snd=nd-sjd # Stoff nicht geübt und nicht bestanden

djb=14     # zum lernen daheim geblieben und bestanden 
djd=1      # zum lernen daheim geblieben und nicht bestanden
dnb=nb-djb # nicht immer zum lernen daheim geblieben und bestanden 
dnd=nd-djd # nicht immer zum lernen daheim geblieben und nicht bestanden

prio_b=nb/ng  # Vorwahrscheinlichkeit fürs Bestehen der Mathematura
prio_d=nd/ng  # Vorwahrscheinlichkeit für Nicht-Bestehen

print('------------------------')
print('Vorwahrscheinlichkeit fur Bestehen='+str(prio_b))
print('Vorwahrscheinlichkeit für Nicht-Bestehen='+str(prio_d))
print('------------------------')

# Vereinbarung der verwendeten Felder

psb,pub,pdb,psd,pud,pdd=[np.zeros(2) for i in range(6)]
lh_b,lh_d,pkomb_b,pkomb_d,pb,pd,prand,index=[np.zeros(8) for i in range(8)] 

# Bedingte Wahrscheinlichkeiten der Einzelmerkmale bei gegebenem Ergebnis
# Häufigkeiten für ...

psb[1]=sjb/nb   # Stoff geübt, wenn bestanden
psb[0]=snb/nb   # Stoff nicht geübt, wenn bestanden   
pub[1]=ujb/nb   # Unterricht immer besucht, wenn bestanden
pub[0]=unb/nb   # Unterricht nicht immer besucht, wenn bestanden
pdb[1]=djb/nb   # immer daheim geblieben, wenn bestanden
pdb[0]=dnb/nb   # nicht immer daheim geblieben, wenn bestanden


psd[1]=sjd/nd   # Stoff geübt, wenn nicht bestanden 
psd[0]=snd/nd   # Stoff nicht geübt, wenn nicht bestanden
pud[1]=ujd/nd   # ...
pud[0]=und/nd
pdd[1]=djd/nd
pdd[0]=dnd/nd

# Ausdrucken der bedingeten Häufigkeiten

print(sjb,snb,nb,psb)
print(ujb,unb,nb,pub)
print(djb,dnb,nb,pdb)

print(sjd,snd,nd,psd)
print(ujd,und,nd,pud)
print(djd,dnd,nd,pdd)

# Berechnung der Posteriori-Wahrscheinlichkeiten mit Vorwissen

n=0
randsumme=0.
for i in range(0,2):                        # Schleife über Merkmal daheim
    for j in range(0,2):                     # Schleife über Merkmal Unterrichtsbesuch 
        for k in range(0,2):                 # Schleife über Merkmal Stoff üben
            lh_b[n]=pdb[i]*pub[j]*psb[k]     # Likelihood für Merkmalskombinationen falls bestanden
            pkomb_b[n]=prio_b*lh_b[n]        # Kombinierte Wahrscheinlichkeit für Vorbereitungsmerkmale und bestanden
            
            lh_d[n]=pdd[i]*pud[j]*psd[k]     # Likelihood für Merkmalskombinationen falls nicht bestanden
            pkomb_d[n]=prio_d*lh_d[n]        # Kombinierte Wahrscheinlichkeit für Vorbereitungsmerkmale und nicht bestanden
            
            prand[n]=pkomb_b[n]+pkomb_d[n]   # Randwahrscheinlichkeiten
            
            pb[n]=pkomb_b[n]/prand[n]        # Posterioriwahrscheinlichkeiten für bestanden bei gegebenen Merkmalskombinationen 
            pd[n]=pkomb_d[n]/prand[n]        # Posterioriwahrscheinlichkeiten für nicht bestanden bei gegebenen Merkmalskombinationen
            
            index[n]=n
            pkd=round(pkomb_d[n]*100,2)      # Umrechnung in Prozent und Runden auf 2 Nachkommastellen
            pkn=round(pkomb_b[n]*100,2)
            pra=round(prand[n]*100,2)
            pdr=round(pd[n]*100,2)
            pbr=round(pb[n]*100,2)
            print(i,j,k,pkd,pkn,pra,pdr,pbr) # Ausdrucken der Ergebnisse (ijk geben duale Nummer der Merkmalskombination)
            randsumme=randsumme+prand[n]     # Randsumme zur Prüfung, wobei sich 1 ergeben sollte
            n=n+1                            # Hochzählen der Merkmalskombinationen

#print(randsumme)            

xleg=('0 = D\'U\'S\'',         # Merkmalskombinationen (nachgestelltes ' bedeutet Negation)
      '1 = D\'U\'S',
      '2 = D\'US\'',
      '3 = D\'US',
      '4 = DU\'S\'',
      '5 = DU\'S',
      '6 = DUS\'',
      '7 = DUS')
#print(xleg)

plt.figure()                       # <------Beginnt ein Bild
ax=plt.subplot(111)                # <------Handle für ein Teildiagramm (hier insgesamt nur ein Bild)
plt.bar(index+0.00,pb*100.,color='b',width=0.25)
plt.bar(index+0.25,pd*100.,color='r',width=0.25)
plt.grid(True)
plt.ylim(0,100)
plt.xlabel("Merkmalskombination")
plt.ylabel("Bestehenswahrscheinlichkeit")
plt.xticks(np.arange(0,8,step=1))
plt.yticks(np.arange(0,110,step=10))
#plt.legend(labels=["bestanden","nicht bestanden"])
plt.title("Naiver Bayes-Schätzer\n für das Bestehen der Mathematik-Matura")
# Vorbereitung für Legende außerhalb des Diagramms
box= ax.get_position()                                      # Holen der Boxgröße, in der das Diagramm gezeichnet wird       
ax.set_position([box.x0,box.y0,box.width*0.7,box.height])  # Reduktion der Breite auf 75% damit Platz für Legende ist 
# Zeichnet Legende rechts oben außerhalb des Diagramms
plt.legend(labels=["bestanden","nicht bestanden",xleg[0]],bbox_to_anchor=(1.05,1.0),loc='upper left')
for i in range(0,8):
    plt.text(9,50-5*i,xleg[i])
plt.show()


            

