###########################################################################
# Python-Programm für ein kleines neuronales Netz mit  3 Neuronen und  ####
# 2 Einganswerten x1 und x2, einer verborgenen Schicht mit 2 Neuronen  ####
# und einer Ausgabeschicht mit einem Neuron
#
# Als Testbeispiel soll das Netz lernen, ob ein Punkt P(x1|x2) oberhalb oder
# unterhalb einer Geraden x2 = k*x1 +d liegt mit x1 und y1-Werten
# im Intervall [-1,1].
#
##########################################################################
import numpy as np               # stellt Mathe-Hilfsprogramme bereit
import matplotlib.pyplot as plt  # stellt Hilfsprogramme für Diagramme bereit
import random                    # stellt Hilfsprogramm für Zufallszahlen bereit

 
# Funktionsunterprogramm für Aktivierungsfunktion im neuronalen Netz
def af(x):
    y=1./(1.+np.exp(-x))           # Funktionswerte 
    return y

# Funktionsunterprogramm für 1. Ableitung der Aktivierungsfunktion 
def afs(x):
    y=af(x)                         # Berechnet Aktivierungsfunktion
    ys=y*(1-y)                      # Ableitung der Aktivierungsfunktion
    return ys

# Funktionsunterprogramm für Zufallszahlen zwischen -1 und +1
def zuza():
    zufall=random.random()          # Zufallszahlen zwischen 0 und 1
    zz=2.*zufall-1.                 # Zufallszahlen zwischen -1 und 1
    return zz

# Funktionsunterprogramm für Grenzkurve zwischen den Gebieten, wo Punkte klassifiziert werden
def grenze(ig,x):
    if(ig==1):      
        y=k*x+d                     # Grenzkurve ist Gerade
    elif(ig==2):
        y=k*x**2+d                  # Grenzkurve ist Parabel
    elif(ig==3):
        y=k*x**3+d                  # Grenzkurve ist Parabel        
    return y

#########################################################################
# Hier beginnt das Programm für Training und Testen des neuronalen Netzes
#########################################################################

# Parameter und Startwerte für die Rechnung
ntr=100000                     # Anzahl der Trainingsdatensätze
nte=500                        # Anzahl der Testdatensätze
nfe=100                        # Anzahl der Werte über die der Fehler gemittelt wird
eta=0.5                        # Lernrate
ig=1                           # Auswahl der Grenzkurve (ig=1: Gerade, ig=2: Parabel)
d=-0.7                         # Achsenabschnitt der Begrenzungsgerade oder y-Verschiebung der Parabel
k=2.                           # Steigung der Begrenzungsgerade oder Koeffizient vo x**2

# Startwerte für die Gewichte
faktor=0.1                     # Faktor der Gewichte auf Zufallswerte zwischen -0.1 und +0.1 beschränkt

v1=zuza()*faktor
v2=zuza()*faktor
v3=zuza()*faktor

w11=zuza()*faktor
w12=zuza()*faktor
w13=zuza()*faktor

w21=zuza()*faktor
w22=zuza()*faktor
w23=zuza()*faktor

print("===========================================================")
print("Zufällige Anfangsgewichte")
print("v1,v2,v3=")
print(v1,v2,v3)
print("w11,w12,w13=")
print(w11,w12,w13)
print("w21,w22,w23=")
print(w21,w22,w23)

# Bereitstellung von Vektoren für die x- und y-Koordinaten der Ergebnisbilder

iwerte_rot=np.zeros(nte)       # Vektor der Länge nte für die x-Achse der Testdaten über Grenzkurve
iwerte_blau=np.zeros(nte)      # Vektor der Länge nte für die x-Achse der Testdaten unter Grenzkurve
ywerte_rot=np.zeros(nte)       # Vektor der Länge nte für die y-Achse der Testdaten über Grenzkurve
ywerte_blau=np.zeros(nte)      # Vektor der Länge nte für die y-Achse der Testdaten unter Grenzkurve

V1=np.zeros(ntr)               # Vektoren der Länge ntr für die Gewichte            '
V2=np.zeros(ntr)
V3=np.zeros(ntr)

W11=np.zeros(ntr)
W12=np.zeros(ntr)
W13=np.zeros(ntr)

W21=np.zeros(ntr)
W22=np.zeros(ntr)
W23=np.zeros(ntr)

DELTA=np.zeros(ntr)             # Vektor der Länge ntr für den Fehler einer einzelnen Rechnung
fehler=np.zeros(ntr)            # Vektor für den gemittelten Fehler
itrain=np.zeros(ntr)            # Vektor für Nummern der Trainingsdatensätze

# Schleife für Trainingsdaten und Korrektur der Gewichte
for i in range(1,ntr):

# Erstelle einen neuen Trainingsdatensatz (x1,x2,t)

    x1=zuza()              # gibt zufälligen Wert für x1 zwischen -1 und 1 
    x2=zuza()              # gibt zufälligen Wert für x2 zwischen -1 und 1 
    if(x2>grenze(ig,x1)):  # prüft, ob x2 oberhalb der Grenzkurve liegt (ig=1: Gerade oder ig=1: Parabel)        
        t=1                # wenn ja, wird t=1 gesetzt
    else:
        t=0                # wenn nein, wird t=0 gesetzt


# Vorwärtspropagation der Eingabedaten durch das Netz

    a1=w11*x1+w12*x2+w13*1.
    a2=w21*x1+w22*x2+w23*1.
    z1=af(a1)
    z2=af(a2)
    b=v1*z1+v2*z2+v3*1.
    y=af(b)

# Fehler

    delta=y-t
    DELTA[i]=delta

# Mittlerer Fehler    
    fsumme=0. 
    jende=np.minimum(i,nfe)
    for j in range(1,jende+1):
        fsumme=fsumme+np.abs(DELTA[i-j+1])
    itrain[i]=i    
    fehler[i]=fsumme/jende
            

# Korrektur der Gewichte
    v1=v1-eta*delta*afs(b)*z1
    v2=v2-eta*delta*afs(b)*z2    
    v3=v3-eta*delta*afs(b)*1.

    w11=w11-eta*delta*afs(b)*v1*afs(a1)*x1
    w12=w12-eta*delta*afs(b)*v1*afs(a1)*x2       
    w13=w13-eta*delta*afs(b)*v1*afs(a1)*1.    

    w21=w21-eta*delta*afs(b)*v2*afs(a2)*x1
    w22=w22-eta*delta*afs(b)*v2*afs(a2)*x2       
    w23=w23-eta*delta*afs(b)*v2*afs(a2)*1.

# Speichern der korrigierten Gewichte für späteres Zeichen in Diagramm 4

    V1[i]=v1
    V2[i]=v2
    V3[i]=v3

    W11[i]=w11
    W12[i]=w12
    W13[i]=w13

    W21[i]=w21
    W22[i]=w22
    W23[i]=w23

print("===========================================================")
print("Verbesserte Gewichte")
print("v1,v2,v3=")
print(v1,v2,v3)
print("w11,w12,w13=")
print(w11,w12,w13)
print("w21,w22,w23=")
print(w21,w22,w23)

##########################################################################################
# Schleife für Test, ob das Netz Punkte unterhalb und oberhalb einer Geraden erkennen kann

# Es werden 4 Ergebnisdiagramme in ein Bild gezeichnet 
# Erstes Diagramm für die vom Netz gefundenen Testdaten

plt.subplot(2,2,1)      #  Argumente von subplot in der Klammer sind (Zeilen, Spalten,Bildnummer)
plt.tight_layout(pad=1) # Befehl für mehr Abstand zwischen den Diagrammen

ir=0                    # Zähler für oberhalb gefundene Punkte startet mit 0
ib=0                    # Zähler für unterhalb gefundene Punkte startet mit 0

for i in range(1,nte):
    
# Erstelle einen neuen Trainingsdatensatz (x1,x2,t)

    x1=zuza()          # gibt zufälligen Wert für x1 zwischen -1 und 1 
    x2=zuza()          # gibt zufälligen Wert für x2 zwischen -1 und 1 
   
# Vorwärtspropagation der Eingabedaten durch das Netz

    a1=w11*x1+w12*x2+w13*1.
    a2=w21*x1+w22*x2+w23*1.
    z1=af(a1)
    z2=af(a2)
    b=v1*z1+v2*z2+v3*1.
    y=af(b)

# Ausgabe der Testdaten x1, x2 und vom Ergebnis y das das neuronale Netz berechnet
   
    print("x1="+str(x1)+" x2="+str(x2)+" y="+str(y))

# Zeichnen der Punkte oberhalb und unterhalb der Grenzkurve mit unterschiedlicher Farbe  

    if(y>0.5):
        plt.plot(x1,x2,"r+")    # Zeichne Testdaten, die das Netz über Grenzkurve findet 
        ir=ir+1                 # Zählen der oberhalb gefundenen Punkte
        iwerte_rot[ir]=ir        
        ywerte_rot[ir]=y        # Speichern der y-Werte oberhalb in Vektor 
    else:
        plt.plot(x1,x2,"b+")    # Zeichne Testdaten, die das Netz unter Grenzkueve findet
        ib=ib+1                 # Zählen der unterhalb gefundenen Punkte
        iwerte_blau[ib]=ib
        ywerte_blau[ib]=y       # Speichern der y-Werte unterhalb in Vektor
        
# Zeichne Grenze zwischen den Gebieten in erstes Diagramm

xgg=np.linspace(-1,1,100)
ygg=grenze(ig,xgg)
plt.plot(xgg,ygg)
plt.ylim(-1,1)
plt.xlabel("x1")
plt.ylabel("x2")
plt.title("Testdaten") 

# Zweites Diagramm für die Netzausgabewerte
    
plt.subplot(2,2,2)
plt.tight_layout(pad=1)
plt.plot(iwerte_rot[1:ir],ywerte_rot[1:ir],"r.",label="oberhalb")
plt.plot(iwerte_blau[1:ib],ywerte_blau[1:ib],"b.",label="unterhalb")
plt.legend(bbox_to_anchor=(1.05,1.0),loc='upper left')
plt.xlabel("Nummer Testdatensatz")
plt.ylabel("y")
plt.title("Netzausgabewerte y")

# Drittes Diagramm für die Änderung des mittleren Fehlers während des Trainings
    
plt.subplot(2,2,3)
plt.tight_layout(pad=1)
plt.plot(itrain,fehler)
plt.xlabel("Nummer Trainingsdatensatz")
plt.ylabel("gemittelter Fehler")
plt.title("Fehlerentwicklung")

# Viertes Diagramm für die Änderung der Gewichte während des Trainings 
    
plt.subplot(2,2,4)

plt.plot(itrain,V1,label="v1")
plt.plot(itrain,V2,label="v2")
plt.plot(itrain,V3,label="v3")
plt.plot(itrain,W11,label="w11")
plt.plot(itrain,W12,label="w12")
plt.plot(itrain,W13,label="w13")
plt.plot(itrain,W23,label="w21")
plt.plot(itrain,W22,label="w22")
plt.plot(itrain,W23,label="w23")
plt.legend(bbox_to_anchor=(1.05,1.0),loc='upper left')  # Plaziert Legende der Kurven außerhalb vom Bild
plt.tight_layout(pad=1)
plt.xlabel("Nummer Trainingsdatensatz ")
plt.ylabel("Gewichte")
plt.title("Korrigierte Gewichte")
plt.show()
   
