# Programm zur Anwendung des SVM-Algorithmus mit Scikit-Learn zur
# Klassifikation von Punkten, die ober- oder unterhalb einer Geraden liegen
#
import matplotlib.pyplot as plt
import numpy as np
from sklearn import svm
import random                   

random.seed(1)

# Funktionsunterprogramm für Zufallszahlen zwischen -1 und 1
def zuza():
    zufall=random.random()
    zz=2.*zufall-1
    return zz

####################################################################
# Hier beginnt das Programm für die Anwendung des SVM-Algorithmus
####################################################################

# Startwerte für die Rechnung
ntr=50      # Anzahl der Trainingsdaten
nte=50      # Anzahl der Testdaten
k=1.8       # Steigung der Grenzkurve
d=-0.5      # Achsenabschnitt der Grenzkurve
n=1         # Grad des Polynoms

X1=np.zeros(ntr)  # Vektor für x-Werte der Trainingsdaten (alle)
X2=np.zeros(ntr)  # Vektor für y-Werte der Trainingsdaten (alle)

X1o=np.zeros(ntr)  # Vektor für x-Werte der Trainingsdaten (oberhalb)
X2o=np.zeros(ntr)  # Vektor für y-Werte der Trainingsdaten

X1u=np.zeros(ntr)  # Vektor für x-Werte der Trainingsdaten (unterhalb)
X2u=np.zeros(ntr)  # Vektor für y-Werte der Trainingsdaten

T=np.zeros(ntr)
X=np.zeros([ntr,2])

io=0
iu=0

# Schleife für Erzeugung der Trainingsdaten

for i in range(0,ntr):
    X1[i]=zuza()                 # Erzeugen eines zuf"alligen x1-Wertes und Speichern im Feld X1
    X2[i]=zuza()                 # Erzeugen eines zuf"alligen x2-Wertes und Speichern im Feld X2
    X[i,0]=X1[i]
    X[i,1]=X2[i]

    if(X2[i] > k*X1[i]**n+d):    # Prüfung auf Lage des Trainingspunktes relativ zur Grenzgeraden / Punkt ist oberhalb 
        T[i]=1                   # Kennzeichnung 1
        X1o[io]=X1[i]            # Speichern in den Feldern X1o, X2o
        X2o[io]=X2[i]
        io=io+1                  # Zählen der Punkte oberhalb
    else:                        # Punkt ist unterhalb
        T[i]=0                   # Kennzeichnung 0
        X1u[iu]=X1[i]            # Speichern in den Feldern X1u, X2u
        X2u[iu]=X2[i]
        iu=iu+1                  # Zählen der Punkte unterhalb

# Grenzkurve
X1G=np.linspace(-1,1,100)
X2G=k*X1G**n+d

#  Modellanpassung mit linearer SVM (https://scikit-learn.org/stable/modules/svm.html)
clf = svm.SVC(kernel="linear", C=1000)
clf.fit(X, T)

# Diagramme
plt.figure()                       # <------Beginnt ein Bild
ax=plt.subplot(111)                # <------Handle für ein Teildiagramm (hier insgesamt nur ein Bild) 
plt.plot(X1o[0:io],X2o[0:io],"rx") # Zeichnet Trainingsdaten über Grenzkurve
plt.plot(X1u[0:iu],X2u[0:iu],"b+") # Zeichnet Trainingsdaten unter Grenzkurve
plt.plot(X1G,X2G,)                 # Zeichnet die Grenzkurve
plt.ylim(-1,1)                     # Begrenzt x2-Achse auf Intervall [-1,1]
plt.grid(True)                     # Zeichnet ein Gitternetz 

# Trenngerade
w = clf.coef_[0]                       # Normalenvektor w 
b = clf.intercept_[0]                  # b Achsenabschnitt
x_p = np.linspace(-1, 1,100)           # x-Koordinatenpunkte zwischen -1 to 1
y_p = -(w[0] / w[1]) * x_p - b / w[1]  # y-Koordinatenpunkte mit Steigung und Achsenabschnitt
plt.plot(x_p, y_p, "r")

print("Normalenvektor w="+str(w))
betrag=(w[0]**2+w[1]**2)**0.5


# Zeichne Support-Vektoren
ax.scatter(
    clf.support_vectors_[:, 0],
    clf.support_vectors_[:, 1],
    s=100,
    linewidth=1,
    facecolors="none",
    edgecolors="k",
)

# Zeichne Begrenzungsgeraden durch die Support-Vektoren (Ränder) 

# 1. Support-Vektor

sv1=np.zeros(2)
sv1[0]=clf.support_vectors_[0, 0]
sv1[1]=clf.support_vectors_[0, 1]

xm1 = np.linspace(-1, 1,100)    
ym1 = -(w[0] / w[1]) * xm1 + sv1[1]+sv1[0]*w[0]/w[1]
plt.plot(xm1, ym1,'r--')

# 2. Supportvektor

sv2=np.zeros(2)
sv2[0]=clf.support_vectors_[2, 0]
sv2[1]=clf.support_vectors_[2, 1]

xm2 = np.linspace(-1, 1,100)    
ym2 = -(w[0] / w[1]) * xm2 + sv2[1]+sv2[0]*w[0]/w[1]
plt.plot(xm2, ym2, 'r--')

###########################################################################
# Testen mit zufälligen Testdaten
###########################################################################

for j in range(0,nte):          # Schleife für alle Testdaten 
    x1t=zuza()                  # zufälliger x1-Wert für Testpunkt 
    x2t=zuza()                  # zufälliger x2-Wert für Testpunkt
    xtest=[[x1t,x2t]]           # als zweidimensionaler Vektor wie es von clf.predict erwartet wird
    ttest=clf.predict(xtest)    # Klassifikationsergebnis
    if(ttest==1):               # Entscheidung, wie der testpunkt zu zeichnen ist
        plt.plot(x1t,x2t,"mo")  # Zeichne Testpunkt mit Farbe magenta für oberhalb
    else:
        plt.plot(x1t,x2t,"co") # Zeichne Testpunkt mit Farbe cyan für unterhalb
        
plt.xlabel("x1-Koordinate")
plt.ylabel("x2-Koordinate")
plt.title("SVM-Algorithmus mit linearem Kernel")
# Vorbereitung für Legende außerhalb des Diagramms
box= ax.get_position()                                      # Holen der Boxgröße, in der das Diagramm gezeichnet wird       
ax.set_position([box.x0,box.y0,box.width*0.75,box.height])  # Reduktion der Breite auf 75% damit Platz für Legende ist 
plt.legend(labels=["Training","Training","Grenzkurve","Trenngerade","Ränder","Ränder","Testen","Testen"],bbox_to_anchor=(1.05,1.0),loc='upper left')
        
# Zusatzinfos    
abstand=sv1@w/betrag+b/betrag
print("Abstand zwischen Support-Vektor 1 und Trenngerade ="+str(abstand))
print("Support-Vektoren (x-Komp./y-Komp.)")
print(clf.support_vectors_[:, 0])
print(clf.support_vectors_[:, 1])
plt.show()
